from copy import deepcopy

import torch


class OUNoise:
    """
    Implements the [Ornstein-Uhlenbeck](https://journals.aps.org/pr/abstract/10.1103/PhysRev.36.823) (OU) process for generating noise. Often used in Reinforcement Learning to encourage exploration.

    Uses the approach discussed in the DDPG paper: [Continuous Control with Deep Reinforcement Learning](https://arxiv.org/abs/1509.02971).

    $$
    X_{t+1} = X_{t} + \\theta (\\mu - X_{t}) + \\sigma \\xi_{t}
    $$

    Where:

    - $X_t$ is the current noise state.
    - $\\theta$ is the mean-reverting factor (controls how quickly the noise returns to $\\mu$).
    - $\\mu$ is the long-term mean.
    - $\\sigma$ is the noise scale.
    - $\\xi_t \\sim \\mathcal{N}(0,1)$ is standard Gaussian noise.
    """

    def __init__(
        self,
        size: int,
        *,
        mu: float = 0.0,
        theta: float = 0.15,
        sigma: float = 0.2,
        device: torch.device = "cpu",
    ) -> None:
        """
        Parameters:
            size (int): size of the sample tensor (e.g., action size)
            mu (float, optional): ($\\mu$) The mean value to which the noise gravitates
            theta (float, optional): ($\\theta$) The speed of mean reversion
            sigma (float, optional): ($\\sigma$) The scale of the random component
            device (torch.device, optional): the device to perform computations on
        """
        self.mu = torch.full((size,), mu, dtype=torch.float32, device=device)
        self.theta = theta
        self.sigma = sigma
        self.device = device

        self.state = None

        self.reset()

    def reset(self) -> None:
        """Resets the noise process to the mean ($\\mu$) value."""
        self.state = deepcopy(self.mu)

    def sample(self) -> torch.Tensor:
        """
        Generates a new noise sample.

        Returns:
            sample (torch.Tensor): A noise sample of the same shape as $\\mu$.
        """
        t = torch.randn(self.state.shape, dtype=torch.float32, device=self.device)
        dx = self.theta * (self.mu - self.state) + self.sigma * t

        self.state += dx
        return self.state



from collections import deque
from dataclasses import dataclass, astuple
import random
from typing import List, Tuple

import torch

from velora.utils.torch import to_tensor, stack_tensor


@dataclass
class Experience:
    """
    Storage container for a single agent experience.

    Parameters:
        state (torch.Tensor): an environment observation
        action (float): agent action taken in the state
        reward (float): reward obtained for taking the action
        next_state (torch.Tensor): a newly generated environment observation
            after performing the action
        done (bool): environment completion status
    """

    state: torch.Tensor
    action: float
    reward: float
    next_state: torch.Tensor
    done: bool

    def __iter__(self) -> Tuple:
        """
        Iteratively unpacks experience instances as tuples.

        Best used with the [`zip()`](https://docs.python.org/3/library/functions.html#zip) method:

        ```python
        batch = [Experience(...), Experience(...), Experience(...)]
        states, actions, rewards, next_states, dones = zip(*batch)

        # ((s1, s2, s3), (a1, a2, a3), (r1, r2, r3), (ns1, ns2, ns3), (d1, d2, d3))
        ```

        Returns:
            exp (Tuple): the experience as a tuple in the form `(state, action, reward, next_state, done)`.
        """
        return iter(astuple(self))


@dataclass
class BatchExperience:
    """
    Storage container for a batch agent experiences.

    Parameters:
        states (torch.Tensor): a batch of environment observations
        actions (torch.Tensor): a batch of agent actions taken in the states
        rewards (torch.Tensor): a batch of rewards obtained for taking the actions
        next_states (torch.Tensor): a batch of newly generated environment
            observations following the actions taken
        dones (torch.Tensor): a batch of environment completion statuses
    """

    states: torch.Tensor
    actions: torch.Tensor
    rewards: torch.Tensor
    next_states: torch.Tensor
    dones: torch.Tensor


class ReplayBuffer:
    """
    A Buffer for storing agent experiences. Used for Off-Policy agents.

    First introduced in Deep RL in the Deep Q-Network paper:
    [Player Atari with Deep Reinforcement Learning](https://arxiv.org/abs/1312.5602).
    """

    def __init__(self, capacity: int, *, device: torch.device | None = None) -> None:
        """
        Parameters:
            capacity (int): the total capacity of the buffer
            device (torch.device, optional): the device to perform computations on
        """
        self.capacity = capacity
        self.buffer = deque(maxlen=capacity)
        self.device = device

    def push(self, exp: Experience) -> None:
        """
        Stores an experience in the buffer.

        Parameters:
            exp (Experience): a single set of experience as an object
        """
        self.buffer.append(exp)

    def sample(self, batch_size: int) -> BatchExperience:
        """
        Samples a random batch of experiences from the buffer.

        Parameters:
            batch_size (int): the number of items to sample

        Returns:
            batch (BatchExperience): an object of samples with the attributes (`states`, `actions`, `rewards`, `next_states`, `dones`).

                All items have the same shape `(batch_size, features)`.
        """
        if len(self.buffer) < batch_size:
            raise ValueError(
                f"Buffer does not contain enough experiences. Available: {len(self.buffer)}, Requested: {batch_size}"
            )

        batch: List[Experience] = random.sample(self.buffer, batch_size)

        states, actions, rewards, next_states, dones = zip(*batch)

        return BatchExperience(
            states=stack_tensor(states, device=self.device),
            actions=to_tensor(actions, device=self.device).unsqueeze(1),
            rewards=to_tensor(rewards, device=self.device).unsqueeze(1),
            next_states=stack_tensor(next_states, device=self.device),
            dones=to_tensor(dones, device=self.device).unsqueeze(1),
        )

    def __len__(self) -> int:
        """
        Gets the current size of the buffer.

        Returns:
            size (int): the current size of the buffer.
        """
        return len(self.buffer)


class RolloutBuffer:
    """
    A Rollout Buffer for storing agent experiences. Used for On-Policy agents.

    Uses a similar implementation to `ReplayBuffer`. However, it must
    be emptied after it is full.
    """

    def __init__(self, capacity: int, *, device: torch.device | None = None) -> None:
        """
        Parameters:
            capacity (int): Maximum rollout length
            device (torch.device, optional): the device to perform computations on
        """
        self.capacity = capacity
        self.buffer = deque(maxlen=capacity)
        self.device = device

    def push(self, exp: Experience) -> None:
        """
        Stores an experience in the buffer.

        Parameters:
            exp (Experience): a single set of experience as an object
        """
        if len(self.buffer) == self.capacity:
            raise BufferError("Buffer full! Use the 'empty()' method first.")

        self.buffer.append(exp)

    def sample(self) -> BatchExperience:
        """
        Returns the entire rollout buffer as a batch of experience.

        Returns:
            batch (BatchExperience): an object of samples with the attributes (`states`, `actions`, `rewards`, `next_states`, `dones`).

                All items have the same shape `(batch_size, features)`.
        """
        if len(self.buffer) == 0:
            raise BufferError("Buffer is empty!")

        states, actions, rewards, next_states, dones = zip(*self.buffer)

        return BatchExperience(
            states=stack_tensor(states, device=self.device),
            actions=to_tensor(actions, device=self.device).unsqueeze(1),
            rewards=to_tensor(rewards, device=self.device).unsqueeze(1),
            next_states=stack_tensor(next_states, device=self.device),
            dones=to_tensor(dones, device=self.device).unsqueeze(1),
        )

    def empty(self) -> None:
        """Empties the buffer."""
        self.buffer.clear()

    def __len__(self) -> int:
        """
        Gets the current size of the buffer.

        Returns:
            size (int): the current size of the buffer.
        """
        return len(self.buffer)



class IncorrectFileTypeError(Exception):
    def __init__(self, message: str = "Incorrect file type provided.") -> None:
        super().__init__(message)
        self.message = message


class RunNotFoundError(Exception):
    def __init__(self, message: str = "Analytics run failed.") -> None:
        super().__init__(message)
        self.message = message



from abc import ABC, abstractmethod
from typing import Any, Self

from pydantic import BaseModel


class Analytics(ABC, BaseModel):
    """An analytics base class."""

    @abstractmethod
    def init(self, *args: Any, **kwargs: Any) -> Any:
        """Starts a new run for the analytics tracker."""
        pass  # pragma: no cover

    @abstractmethod
    def log(self, metrics: dict[str, Any], **kwargs: Any) -> None:
        """Logs metrics to the run."""
        pass  # pragma: no cover

    @abstractmethod
    def finish(self) -> None:
        """Marks the run as finished, uploads final data, and resets run instance to None."""
        pass  # pragma: no cover


class NullAnalytics(Analytics):
    """An empty analytics tracker for disabling analytics tracking."""

    def init(self, *args: Any, **kwargs: Any) -> Self:
        return self

    def log(self, metrics: dict[str, Any], *args: Any, **kwargs: Any) -> None:
        pass

    def finish(self) -> None:
        pass



from typing import Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F


class NCPLiquidCell(nn.Module):
    """
    A Neural Circuit Policy (NCP) Liquid Time-Constant (LTC) cell.

    The LTC cell follows the closed-form continuous-depth
    (CFC; Equation 10) solution from the paper:
    [Closed-form Continuous-time Neural Models](https://arxiv.org/abs/2106.13898).

    Plus, it follows an Ordinary Neural Circuit (ONC) approach from this paper:
    [Reinforcement Learning with Ordinary Neural Circuits](https://proceedings.mlr.press/v119/hasani20a.html).

    Equation:
    $$
    x(t) =
        \\sigma(-f(x, I, θ_f), t) \\; g(x, I, θ_g)
        + \\left[ 1 - \\sigma(-[\\;f(x, I, θ_f)\\;]\\;t) \\right] \\; h(x, I, θ_h)
    $$
    """

    def __init__(
        self,
        in_features: int,
        n_hidden: int,
        mask: torch.Tensor,
        *,
        device: torch.device | None = None,
    ) -> None:
        """
        Parameters:
            in_features (int): number of input nodes.
            n_hidden (int): number of hidden nodes.
            mask (torch.Tensor): a matrix of sparse connections
                usually containing a combination of `[-1, 1, 0]`.
            device (torch.device, optional): the device to load tensors on.
        """

        super().__init__()

        self.in_features = in_features
        self.n_hidden = n_hidden
        self.head_size = n_hidden + in_features
        self.device = device

        # Absolute to maintain masking (-1 -> 1)
        self.sparsity_mask = nn.Parameter(
            self._prep_mask(mask.to(device)),
            requires_grad=False,
        )

        self.tanh = nn.Tanh()  # Bounded: [-1, 1]
        self.sigmoid = nn.Sigmoid()  # Bounded: [0, 1]

        self.g_head = nn.Linear(self.head_size, n_hidden, device=device)
        self.h_head = nn.Linear(self.head_size, n_hidden, device=device)

        # LTC heads (f)
        self.f_head_to_g = nn.Linear(self.head_size, n_hidden, device=device)
        self.f_head_to_h = nn.Linear(self.head_size, n_hidden, device=device)

        # Hidden state projection
        self.proj = nn.Linear(self.head_size, n_hidden, device=device)

    def _prep_mask(self, mask: torch.Tensor) -> torch.Tensor:
        """
        Utility method. Preprocesses mask to match head size.

        !!! note "Performs three operations"

            1. Adds a padded matrix of 1s to end of mask in shape
                `(n_extras, n_extras)` where `n_extras=mask.shape[1]`
            2. Transposes mask from col matrix -> row matrix
            3. Gets the absolute values of the mask (swapping `-1 -> 1`)

        Parameters:
            mask (torch.Tensor): weight sparsity mask.

        Returns:
            mask (torch.Tensor): an updated mask.
        """
        n_extras = mask.shape[1]
        extra_nodes = torch.ones((n_extras, n_extras), device=self.device)
        mask = torch.concatenate([mask, extra_nodes])
        return torch.abs(mask.T)

    def _sparse_head(self, x: torch.Tensor, head: nn.Linear) -> torch.Tensor:
        """
        Utility method. Computes the output for a sparsity mask head.

        Parameters:
            x (torch.Tensor): layer inputs.
            head (nn.Linear): linear head to use with sparse connections.

        Returns:
            y_pred (torch.Tensor): sparse head prediction.
        """
        return F.linear(
            x.to(torch.float32),
            head.weight * self.sparsity_mask,
            head.bias,
        )

    def _new_hidden(
        self, x: torch.Tensor, g_out: torch.Tensor, h_out: torch.Tensor
    ) -> torch.Tensor:
        """
        Helper method. Computes the new hidden state.

        Parameters:
            x (torch.Tensor): input values.
            g_out (torch.Tensor): g_head output.
            h_out (torch.Tensor): h_head output.

        Returns:
            hidden (torch.Tensor): a new hidden state
        """
        g_head = self.tanh(g_out)  # g(x, I, θ_g)
        h_head = self.tanh(h_out)  # h(x, I, θ_h)

        fh_g = self._sparse_head(x, self.f_head_to_g)
        fh_h = self._sparse_head(x, self.f_head_to_h)

        gate_out = self.sigmoid(fh_g + fh_h)  # [1 - σ(-[f(x, I, θf)], t)]
        f_head = 1.0 - gate_out  # σ(-f(x, I, θf), t)

        return g_head * f_head + gate_out * h_head

    def forward(
        self, x: torch.Tensor, hidden: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Performs a forward pass through the cell.

        Parameters:
            x (torch.Tensor): input values.
            hidden (torch.Tensor): current hidden state.

        Returns:
            y_pred (torch.Tensor): the cell prediction.
            h_state (torch.Tensor): the hidden state.
        """
        x, hidden = x.to(self.device), hidden.to(self.device)
        x = torch.cat([x, hidden], dim=1)

        g_out = self._sparse_head(x, self.g_head)
        h_out = self._sparse_head(x, self.h_head)

        new_hidden = self._new_hidden(x, g_out, h_out)
        y_pred = self._sparse_head(x, self.proj) + new_hidden
        return y_pred, new_hidden



from pydantic import BaseModel, Field

from velora.analytics.base import Analytics, NullAnalytics
from velora.analytics.wandb import WeightsAndBiases

__all__ = [
    "Analytics",
    "WeightsAndBiases",
    "NullAnalytics",
]


class EpisodeStats(BaseModel):
    """
    Stores episode statistics as a model.

    Args:
        reward (float): cumulative reward
        size (int): episode length
        time_elapsed (float): elapsed time since beginning of episode
    """

    reward: float = Field(..., alias="r")
    size: int = Field(..., alias="l")
    time_elapsed: float = Field(..., alias="t")



from typing import Any
from pydantic import PrivateAttr

import wandb
from wandb.sdk.wandb_run import Run

from velora.analytics.base import Analytics
from velora.exc import RunNotFoundError


class WeightsAndBiases(Analytics):
    """A class dedicated to working with [Weights and Biases](https://wandb.ai/)."""

    _run: Run | None = PrivateAttr(None)

    def model_post_init(self, __context: Any) -> None:
        wandb.login()

    @property
    def run(self) -> Run:
        """Returns the active W&B run."""
        return self._run

    def init(
        self, project_name: str, run_name: str, config: dict[str, Any], **kwargs: Any
    ) -> None:
        """
        Starts a new run to track and log W&B.

        Args:
            project_name (str): The name of the project the run belongs to
            run_name (str): The unique display name for the run
            config (dict[str, Any]): A dictionary of inputs and hyperparameters used during the run
        """
        self._run = wandb.init(
            project=project_name,
            name=run_name,
            config=config,
            monitor_gym=True,
            **kwargs,
        )

    def log(self, metrics: dict[str, Any], **kwargs: Any) -> None:
        """
        Logs metrics to the Weights and Biases run.

        Args:
            metrics (dict[str, Any]): a dictionary of metrics to log with names and values
        """
        if not isinstance(self._run, Run):
            raise RunNotFoundError(
                "No run instance found. Have you called the 'init' method?"
            )

        self._run.log(data=metrics, **kwargs)

    def finish(self, **kwargs: Any) -> None:
        """Marks the run as finished, uploads final data, and resets run instance to None."""
        if not isinstance(self._run, Run):
            raise RunNotFoundError(
                "No run instance found. Have you called the 'init' method?"
            )

        self._run.finish(**kwargs)
        self._run = None



from collections import OrderedDict
from typing import Optional, Tuple

import torch
import torch.nn as nn

from velora.models.lnn.cell import NCPLiquidCell
from velora.wiring import Wiring


class LiquidNCPNetwork(nn.Module):
    """
    A Liquid Neural Circuit Policy (NCP) Network with three layers:

    1. Inter (input)
    2. Command (hidden)
    3. Motor (output)

    Each layer is a `NCPLiquidCell`.

    !!! note "Decision nodes"

        `inter` and `command` neurons are automatically calculated using:

        ```python
        command_neurons = max(int(0.4 * n_neurons), 1)
        inter_neurons = n_neurons - command_neurons
        ```
    """

    def __init__(
        self,
        in_features: int,
        n_neurons: int,
        out_features: int,
        *,
        sparsity_level: float = 0.5,
        device: torch.device | None = None,
    ) -> None:
        """
        Parameters:
            in_features (int): number of inputs (sensory nodes)
            n_neurons (int): number of decision nodes (inter and command nodes).
            out_features (int): number of out features (motor nodes)
            sparsity_level (float, optional): controls the connection sparsity
                between neurons.

                Must be a value between `[0.1, 0.9]` -

                - When `0.1` neurons are very dense.
                - When `0.9` they are very sparse.

            device (torch.device, optional): the device to load tensors on.
        """
        super().__init__()

        self.in_features = in_features
        self.n_neurons = n_neurons
        self.out_features = out_features
        self.device = device

        self.n_units = n_neurons + out_features  # inter + command + motor

        self._wiring = Wiring(
            in_features,
            n_neurons,
            out_features,
            sparsity_level=sparsity_level,
        )
        self._masks, self._counts = self._wiring.data()

        names = ["inter", "command", "motor"]
        layers = [
            NCPLiquidCell(
                in_features,
                self._counts.inter,
                self._masks.inter,
            ).to(device),
            NCPLiquidCell(
                self._counts.inter,
                self._counts.command,
                self._masks.command,
            ).to(device),
            NCPLiquidCell(
                self._counts.command,
                self._counts.motor,
                self._masks.motor,
            ).to(device),
        ]
        self.layers = OrderedDict([(name, layer) for name, layer in zip(names, layers)])

        self.ncp = nn.Sequential(self.layers)
        self._out_sizes = [layer.n_hidden for layer in self.layers.values()]

    def _ncp_forward(
        self, x: torch.Tensor, hidden: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Performs a single timestep through the network layers.

        Splits the hidden state into respective chunks for each layer
        (`out_features`) to maintain their own independent hidden state dynamics.

        Then, merges them together to create a new hidden state.

        Parameters:
            x (torch.Tensor): the current batch of data for the timestep with
                shape: `(batch_size, features)`
            hidden (torch.Tensor): the current hidden state

        Returns:
            y_pred (torch.Tensor): the network prediction.
            new_h_state (torch.Tensor): the merged hidden state from all layers (updated state memory).
        """
        h_state = torch.split(hidden, self._out_sizes, dim=1)

        new_h_state = []
        inputs = x

        # Handle layer independence
        for i, layer in enumerate(self.layers.values()):
            y_pred, h = layer(inputs, h_state[i])
            inputs = y_pred  # (batch_size, layer_out_features)
            new_h_state.append(h)

        new_h_state = torch.cat(new_h_state, dim=1)  # (batch_size, n_units)
        return y_pred, new_h_state

    def forward(
        self, x: torch.Tensor, h_state: Optional[torch.Tensor] = None
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Performs a forward pass through the network.

        Parameters:
            x (torch.Tensor): an input tensor of shape: `(batch_size, features)`.

                - `batch_size` the number of samples per timestep.
                - `features` the features at each timestep (e.g.,
                image features, joint coordinates, word embeddings, raw amplitude
                values).
            h_state (torch.Tensor, optional): initial hidden state of the RNN with
                shape: `(batch_size, n_units)`.

                - `batch_size` the number of samples.
                - `n_units` the total number of hidden neurons
                    (`n_neurons + out_features`).

        Returns:
            y_pred (torch.Tensor): the network prediction. When `batch_size=1`. Out shape is `(out_features)`. Otherwise, `(batch_size, out_features)`.
            h_state (torch.Tensor): the final hidden state. Output shape is `(batch_size, n_units)`.
        """
        if x.dim() != 2:
            raise ValueError(
                f"Unsupported dimensionality: '{x.shape=}'. Should be 2 dimensional with: '(batch_size, features)'."
            )

        x = x.to(torch.float32).to(self.device)

        batch_size, features = x.size()

        if h_state is None:
            h_state = torch.zeros((batch_size, self.n_units), device=self.device)

        # Batch -> (batch_size, out_features)
        y_pred, h_state = self._ncp_forward(x, h_state.to(self.device))

        # Single item -> (out_features)
        if y_pred.shape[0] == 1:
            y_pred = y_pred.squeeze(0)

        # h_state -> (batch_size, n_units)
        return y_pred, h_state



from copy import deepcopy
from typing import Tuple, Type

from velora.buffer import BatchExperience, Experience, ReplayBuffer
from velora.gym import add_core_env_wrappers
from velora.models import LiquidNCPNetwork
from velora.utils.torch import soft_update

import torch
import torch.nn as nn
import torch.optim as optim

import gymnasium as gym
import numpy as np

from velora.noise import OUNoise


class DDPGActor(nn.Module):
    """
    A Liquid NCP Actor Network for the DDPG algorithm.
    """

    def __init__(
        self,
        num_obs: int,
        n_neurons: int,
        num_actions: int,
        *,
        device: torch.device | None = None,
    ):
        """
        Parameters:
            num_obs (int): the number of input observations
            n_neurons (int): the number of hidden neurons
            num_actions (int): the number of actions
            device (torch.device, optional): the device to perform computations on
        """

        super().__init__()

        self.ncp = LiquidNCPNetwork(
            in_features=num_obs,
            n_neurons=n_neurons,
            out_features=num_actions,
            device=device,
        )

    def forward(
        self, obs: torch.Tensor, hidden: torch.Tensor | None = None
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Performs a forward pass through the network.

        Parameters:
            obs (torch.Tensor): the batch of state observations
            hidden (torch.Tensor, optional): the hidden state

        Returns:
            actions (torch.Tensor): the action predictions.
            hidden (torch.Tensor): the new hidden state.
        """
        actions, new_hidden = self.ncp(obs, hidden)
        scaled_actions = torch.tanh(actions)  # Bounded: [-1, 1]
        return scaled_actions, new_hidden


class DDPGCritic(nn.Module):
    """
    A Liquid NCP Critic Network for the DDPG algorithm.
    """

    def __init__(
        self,
        num_obs: int,
        n_neurons: int,
        num_actions: int,
        *,
        device: torch.device | None = None,
    ):
        """
        Parameters:
            num_obs (int): the number of input observations
            n_neurons (int): the number of hidden neurons
            num_actions (int): the number of actions
            device (torch.device, optional): the device to perform computations on
        """
        super().__init__()

        self.ncp = LiquidNCPNetwork(
            in_features=num_obs + num_actions,
            n_neurons=n_neurons,
            out_features=1,  # Q-value output
            device=device,
        )

    def forward(
        self,
        obs: torch.Tensor,
        actions: torch.Tensor,
        hidden: torch.Tensor | None = None,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Performs a forward pass through the network.

        Parameters:
            obs (torch.Tensor): the batch of state observations
            actions (torch.Tensor): the batch of actions
            hidden (torch.Tensor, optional): the hidden state

        Returns:
            q_values (torch.Tensor): the Q-Value predictions.
            hidden (torch.Tensor): the new hidden state.
        """
        inputs = torch.cat([obs, actions], dim=-1)

        q_values, new_hidden = self.ncp(inputs, hidden)
        return q_values, new_hidden


class LiquidDDPG:
    """
    A Liquid variant of the Deep Deterministic Policy Gradient (DDPG)
    algorithm from the paper: [Continuous Control with Deep Reinforcement Learning](https://arxiv.org/abs/1509.02971).

    !!! note "Decision nodes"

        `inter` and `command` neurons are automatically calculated using:

        ```python
        command_neurons = max(int(0.4 * n_neurons), 1)
        inter_neurons = n_neurons - command_neurons
        ```
    """

    def __init__(
        self,
        state_dim: int,
        n_neurons: int,
        action_dim: int,
        *,
        optim: Type[optim.Optimizer] = optim.Adam,
        buffer_size: int = 100_000,
        actor_lr: float = 1e-4,
        critic_lr: float = 1e-3,
        device: torch.device | None = None,
    ) -> None:
        """
        Parameters:
            state_dim (int): number of inputs (sensory nodes)
            n_neurons (int): number of decision nodes (inter and command nodes).
            action_dim (int): number of outputs (motor nodes)
            optim (Type[torch.optim.Optimizer], optional): the type of `PyTorch`
                optimizer to use
            buffer_size (int, optional): the maximum size of the ReplayBuffer
            actor_lr (float, optional): the actor optimizer learning rate
            critic_lr (float, optional): the critic optimizer learning rate
            device (torch.device, optional): the device to perform computations on
        """
        self.state_dim = state_dim
        self.n_neurons = n_neurons
        self.action_dim = action_dim
        self.device = device

        self.actor = DDPGActor(
            self.state_dim,
            self.n_neurons,
            self.action_dim,
            device=self.device,
        ).to(self.device)

        self.critic = DDPGCritic(
            self.state_dim,
            self.n_neurons,
            self.action_dim,
            device=self.device,
        ).to(self.device)

        self.actor_target = deepcopy(self.actor)
        self.critic_target = deepcopy(self.critic)

        # Freeze target networks
        for p in self.actor_target.parameters():
            p.requires_grad = False
        for p in self.critic_target.parameters():
            p.requires_grad = False

        self.actor_optim = optim(self.actor.parameters(), lr=actor_lr)
        self.critic_optim = optim(self.critic.parameters(), lr=critic_lr)

        self.loss = nn.MSELoss()
        self.buffer = ReplayBuffer(capacity=buffer_size, device=device)
        self.noise = OUNoise(action_dim, device=device)

    def _update_target_networks(self, tau: float) -> None:
        """
        Helper method. Performs a soft update on the target networks.

        Parameters:
            tau (float): a soft decay coefficient for updating the target network
                weights
        """
        soft_update(self.actor, self.actor_target, tau)
        soft_update(self.critic, self.critic_target, tau)

    def _update_critic(self, batch: BatchExperience, gamma: float) -> float:
        """
        Helper method. Performs a Critic Network update.

        Parameters:
            batch (BatchExperience): an object containing a batch of experience
                with `(states, actions, rewards, next_states, dones)` from the
                buffer
            gamma (float): the reward discount factor

        Returns:
            critic_loss (float): the Critic's loss value.
        """
        with torch.no_grad():
            next_states = batch.next_states
            next_actions, _ = self.actor_target(next_states)
            target_q, _ = self.critic_target(next_states, next_actions)
            target_q = batch.rewards + (1 - batch.dones) * gamma * target_q

        current_q, _ = self.critic(batch.states, batch.actions)
        critic_loss: torch.Tensor = self.loss(current_q, target_q)

        self.critic_optim.zero_grad()
        critic_loss.backward()
        self.critic_optim.step()

        return critic_loss.item()

    def _update_actor(self, states: torch.Tensor) -> float:
        """
        Helper method. Performs an Actor Network update.

        Parameters:
            states (torch.Tensor): a batch of state experiences from the buffer

        Returns:
            actor_loss (float): the Actor's loss value.
        """
        next_actions, _ = self.actor(states)
        actor_q, _ = self.critic(states, next_actions)
        actor_loss: torch.Tensor = -actor_q.mean()

        self.actor_optim.zero_grad()
        actor_loss.backward()
        self.actor_optim.step()

        return actor_loss.item()

    def _train_step(self, batch_size: int, gamma: float) -> Tuple[float, float]:
        """
        Helper method. Performs a single training step.

        Parameters:
            batch_size (int): number of samples in a batch
            gamma (float): the reward discount factor

        Returns:
            critic_loss (float): the critic loss.
            actor_loss (float): the actor loss.
        """
        if len(self.buffer) < batch_size:
            return

        batch: BatchExperience = self.buffer.sample(batch_size)

        critic_loss = self._update_critic(batch, gamma)
        actor_loss = self._update_actor(batch.states)

        return critic_loss, actor_loss

    def train(
        self,
        env: gym.Env,
        batch_size: int,
        *,
        n_episodes: int = 1000,
        max_steps: int = 1000,
        noise_scale: float = 0.1,
        gamma: float = 0.99,
        tau: float = 0.005,
        output_count: int = 100,
    ) -> None:
        """
        Trains the agent on a Gymnasium environment using a `ReplayBuffer`.

        Parameters:
            env (gym.Env): the Gymnasium environment to train on
            batch_size (int): the number of features in a single batch
            n_episodes (int, optional): the total number of episodes to train for
            max_steps (int, optional): the total number of steps per episode
            noise_scale (float, optional): the exploration noise added when
                selecting an action
            gamma (float, optional): the reward discount factor
            tau (float, optional): the soft update factor used to slowly update
                the target networks
            output_count (int, optional): the episodic rate for displaying
                information to the console

        Returns:
            results (TrainResults): an object containing training results.
        """
        if not isinstance(env.action_space, gym.spaces.Box):
            raise EnvironmentError(
                f"Invalid '{env.action_space=}'. Must be 'gym.spaces.Box'."
            )

        env = add_core_env_wrappers(env, self.device)

        episode_rewards = []
        training_started = False

        print(f"{batch_size=}, getting buffer samples.")
        for i_ep in range(n_episodes):
            state, _ = env.reset()

            episode_reward = 0
            critic_losses, actor_losses = [], []
            actor_hidden = None

            for i_step in range(max_steps):
                action, actor_hidden = self.predict(
                    state,
                    actor_hidden,
                    noise_scale=noise_scale,
                )
                next_state, reward, terminated, truncated, _ = env.step(action)
                done = terminated or truncated

                self.buffer.push(
                    Experience(state, action.item(), reward, next_state, done),
                )

                if len(self.buffer) >= batch_size:
                    if not training_started:
                        print("Buffer warmed. Starting training...")
                        training_started = True

                    critic_loss, actor_loss = self._train_step(batch_size, gamma)
                    self._update_target_networks(tau)

                    critic_losses.append(critic_loss)
                    actor_losses.append(actor_loss)

                state = next_state
                episode_reward += reward

                if done:
                    break

            episode_rewards.append(episode_reward)

            if training_started and (i_ep + 1) % output_count == 0:
                avg_reward = np.mean(episode_rewards[-output_count:])
                avg_critic_loss = np.mean(critic_losses)
                avg_actor_loss = np.mean(actor_losses)

                print(
                    f"Episode: {i_ep + 1}/{n_episodes}, "
                    f"Avg Reward: {avg_reward:.2f}, "
                    f"Critic Loss: {avg_critic_loss:.2f}, "
                    f"Actor Loss: {avg_actor_loss:.2f}"
                )

        return episode_rewards

    def predict(
        self,
        state: torch.Tensor,
        hidden: torch.Tensor = None,
        *,
        noise_scale: float = 0.1,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Makes an action prediction using the Actor network with exploration noise.

        Parameters:
            state (torch.Tensor): the current state
            hidden (torch.Tensor, optional): the current hidden state
            noise_scale (float, optional): the exploration noise added when
                selecting an action

        Returns:
            action (torch.Tensor): the action prediction on the given state
            hidden (torch.Tensor): the Actor networks new hidden state
        """
        self.actor.eval()
        with torch.no_grad():
            action, hidden = self.actor(state.unsqueeze(0), hidden)

            if noise_scale > 0:
                # Exploration noise
                noise = self.noise.sample() * noise_scale
                action = torch.clamp(action + noise, min=-1, max=1)

        self.actor.train()
        return action.flatten(), hidden



from velora.models import LiquidNCPNetwork

import torch
import torch.nn as nn


class LiquidPPO(nn.Module):
    """
    A Liquid Network variant of the Proximal Policy Optimization (PPO)
    algorithm.

    Parameters:
        in_features (int): number of inputs (sensory nodes)
        n_neurons (int): number of decision nodes (inter and command nodes).
            Nodes are set automatically based on the following:
            ```python
            command_neurons = max(int(0.4 * n_neurons), 1)
            inter_neurons = n_neurons - command_neurons
            ```
        out_features (int): number of out features (motor nodes)
        sparsity_level (float, optional): controls the connection sparsity
            between neurons. Must be a value between `[0.1, 0.9]`. When `0.1` neurons are very dense, when `0.9` they are very sparse. Default
            is '0.5'
        device (torch.device, optional): the device to load `torch.Tensors` onto.
            Default is 'None'
    """

    def __init__(
        self,
        in_features: int,
        n_neurons: int,
        out_features: int,
        *,
        sparsity_level: float = 0.5,
        device: torch.device | None = None,
    ) -> None:
        super().__init__()

        self.in_features = in_features
        self.n_neurons = n_neurons
        self.out_features = out_features

        self.actor = LiquidNCPNetwork(
            in_features,
            n_neurons,
            out_features,
            sparsity_level=sparsity_level,
            device=device,
        ).to(device)
        self.critic = LiquidNCPNetwork(
            in_features,
            n_neurons,
            out_features,
            sparsity_level=sparsity_level,
            device=device,
        ).to(device)



import torch
import numpy as np


def set_seed(value: int) -> None:
    """
    Sets the random seed for the `PyTorch` and `NumPy` packages.

    Parameters:
        value (int): the seed value
    """
    torch.manual_seed(value)
    np.random.seed(value)


def set_device(device: str = "auto") -> torch.device:
    """
    Sets the `PyTorch` device dynamically.

    Parameters:
        device (str, optional): the name of the device to perform computations on.

            When `auto`:

            - Set to `cuda:0`, if available.
            - Else, `cpu`.
    """
    if device == "auto":
        device = "cuda:0" if torch.cuda.is_available() else "cpu"

    print(f"Device set to '{device}'.")
    return torch.device(device)



from functools import reduce
import re
from typing import Callable, Literal, List, Tuple

import gymnasium as gym
from gymnasium import spaces
from gymnasium.wrappers import RecordEpisodeStatistics
from gymnasium.wrappers.numpy_to_torch import NumpyToTorch

import torch


WrapperType = List[gym.Wrapper | gym.vector.VectorWrapper | Callable]

ContinuousGymName = Literal[
    "MountainCarContinuous-v0",
    "Pendulum-v1",
    "LunarLanderContinuous-v3",
    "BipedalWalker-v3",
    "CarRacing-v3",
    "Reacher-v5",
    "Pusher-v5",
    "InvertedPendulum-v5",
    "InvertedDoublePendulum-v5",
    "HalfCheetah-v5",
    "Hopper-v5",
    "Swimmer-v5",
    "Walker2d-v5",
    "Ant-v5",
    "Humanoid-v5",
    "HumanoidStandup-v5",
]


def wrap_gym_env(
    env: gym.Env | str | ContinuousGymName,
    wrappers: List[gym.Wrapper | gym.vector.VectorWrapper | Callable],
) -> gym.Env:
    """
    Creates a new [Gymnasium](https://gymnasium.farama.org/) environment with
    one or more [gymnasium.Wrappers](https://gymnasium.farama.org/api/wrappers/table/) applied.

    Parameters:
        env (gymnasium.Env | str): a name of a Gymnasium environment or the
            environment itself to wrap
        wrappers (List[gym.Wrapper | gym.vector.VectorWrapper | functools.partial]): a list of wrapper classes or partially applied wrapper functions

    Examples:
        A Gymnasium environment with normalization and reward clipping:
        ```python
        from functools import partial

        from gymnasium.wrappers import (
            NormalizeObservation,
            NormalizeReward,
            RecordEpisodeStatistics,
            ClipReward,
        )
        from gymnasium.wrappers.numpy_to_torch import NumpyToTorch

        import torch

        env = wrap_gym_env("InvertedPendulum-v5", [
            partial(NormalizeObservation, epsilon=1e-8),
            partial(NormalizeReward, gamma=0.99, epsilon=1e-8),
            partial(ClipReward, max_reward=10.0),
            # RecordEpisodeStatistics,
            # partial(NumpyToTorch, device=torch.device("cuda"))
        ])
        ```

    Returns:
        env (gymnasium.Env): The wrapped environment
    """
    if isinstance(env, str):
        env = gym.make(env)

    def apply_wrapper(env: gym.Env, wrapper: WrapperType) -> gym.Env:
        return wrapper(env)

    return reduce(apply_wrapper, wrappers, env)


def add_core_env_wrappers(env: gym.Env, device: torch.device) -> gym.Env:
    """
    Wraps a [Gymnasium](https://gymnasium.farama.org/) environment with the following (in order) if not already applied:

    - [RecordEpisodeStatistics](https://gymnasium.farama.org/api/wrappers/misc_wrappers/#gymnasium.wrappers.RecordEpisodeStatistics)
    - [NumpyToTorch](https://gymnasium.farama.org/api/wrappers/misc_wrappers/#gymnasium.wrappers.NumpyToTorch)

    Used in all pre-built algorithms.

    Parameters:
        env (gym.Env): the Gymnasium environment
        device (torch.device): the PyTorch device to perform computations on
    """
    if not isinstance(env, RecordEpisodeStatistics):
        env = RecordEpisodeStatistics(env)

    if not isinstance(env, NumpyToTorch):
        env = NumpyToTorch(env, device=device)

    return env


def make_wrapped_vec_env(
    name: str,
    *,
    n_envs: int = 3,
    epsilon: float = 1e-8,
    max_reward: float = 10.0,
    gamma: float = 0.99,
    device: str = "cpu",
) -> gym.vector.SyncVectorEnv:
    """
    Creates a vectorized [Gymnasium environment](https://gymnasium.farama.org/api/vector/) with predefined wrappers.

    Wrappers included:

    - [NormalizeObservation](https://gymnasium.farama.org/api/vector/wrappers/#gymnasium.wrappers.vector.NormalizeObservation)
    - [NormalizeReward](https://gymnasium.farama.org/api/vector/wrappers/#gymnasium.wrappers.vector.NormalizeReward)
    - [RecordEpisodeStatistics](https://gymnasium.farama.org/api/vector/wrappers/#gymnasium.wrappers.vector.RecordEpisodeStatistics)
    - [ClipReward](https://gymnasium.farama.org/api/vector/wrappers/#gymnasium.wrappers.vector.ClipReward)
    - [NumpyToTorch](https://gymnasium.farama.org/api/vector/wrappers/#gymnasium.wrappers.vector.NumpyToTorch)

    Parameters:
        name (str): the name of the environment
        n_envs (int, optional): the number of parallel environments
        epsilon (float, optional): stability parameter for normalization scaling.
            Used in the `NormalizeObservation` and `NormalizeReward` wrappers
        max_reward (float, optional): the max absolute value for discounted return.
            Used in the `ClipReward` wrapper
        gamma (float, optional): the discount factor used in `NormalizeReward`
            wrapper. Applies to the exponential moving average
        device (str, optional): the type of CUDA device to load onto.
            Used in the `NumpyToTorch` wrapper
    """
    from gymnasium.wrappers.vector import (
        NormalizeObservation,
        NormalizeReward,
        RecordEpisodeStatistics,
        ClipReward,
    )
    from gymnasium.wrappers.vector.numpy_to_torch import NumpyToTorch

    envs = gym.make_vec(
        name,
        num_envs=n_envs,
        vectorization_mode="sync",
    )
    envs = NormalizeObservation(envs, epsilon=epsilon)
    envs = NormalizeReward(envs, gamma=gamma, epsilon=epsilon)
    envs = ClipReward(envs, max_reward=max_reward)
    envs = RecordEpisodeStatistics(envs)
    envs = NumpyToTorch(envs, device=torch.device(device))
    return envs


def get_action_bounds(action_space: spaces.Box) -> Tuple[float, float]:
    """
    Gets the action bounds from a [Box](https://gymnasium.farama.org/api/spaces/fundamental/#gymnasium.spaces.Box) action space.

    Returns:
        bounds (Tuple[float, float]): the `(low, high)` action thresholds.
    """
    if not isinstance(action_space, spaces.Box):
        raise ValueError(f"'{action_space}' action space is not supported.")

    return action_space.low.item(), action_space.high.item()


def get_latest_env_names() -> List[str]:
    """
    Returns a list of the latest [Gymnasium](https://gymnasium.farama.org/) environment names.

    Returns:
        gym_names (List[str]): a list of names for all latest versions.
    """
    env_dict = {}

    for env in gym.envs.registry.keys():
        match = re.match(r"(.+)-v(\d+)", env)
        if match:
            base_name, version = match.groups()
            version = int(version)

            # Keep only the latest version
            if base_name not in env_dict or version > env_dict[base_name]:
                env_dict[base_name] = version

    # Reconstruct the latest version environments
    latest_envs = [f"{name}-v{version}" for name, version in env_dict.items()]
    return latest_envs



from velora.utils.core import set_device, set_seed
from velora.utils.torch import to_tensor, stack_tensor, soft_update, hard_update


__all__ = [
    "set_device",
    "set_seed",
    "to_tensor",
    "stack_tensor",
    "soft_update",
    "hard_update",
]



from typing import Any, List

import torch
import torch.nn as nn


def to_tensor(
    items: List[Any],
    *,
    dtype: torch.dtype = torch.float32,
    device: torch.device | None = None,
) -> torch.Tensor:
    """
    Converts a list of items to a Tensor, then:

    1. Converts it to a specific `dtype`
    2. Loads it onto `device`

    Parameters:
        items (List[Any]): a list of items of any type
        dtype (torch.dtype, optional): the data type for the tensor
        device (torch.device, optional): the device to perform computations on
    """
    return torch.tensor(items).to(dtype).to(device)


def stack_tensor(
    items: List[torch.Tensor],
    *,
    dtype: torch.dtype = torch.float32,
    device: torch.device | None = None,
) -> torch.Tensor:
    """
    Stacks a list of tensors together, then:

    1. Converts it to a specific `dtype`
    2. Loads it onto `device`

    Parameters:
        items (List[torch.Tensor]): a list of torch.Tensors full of items
        dtype (torch.dtype, optional): the data type for the tensor
        device (torch.device, optional): the device to perform computations on
    """
    return torch.stack(items).to(dtype).to(device)


def soft_update(source: nn.Module, target: nn.Module, tau: float = 0.005) -> None:
    """
    Performs a soft parameter update between two PyTorch Networks.

    Parameters:
        source (nn.Module): the source network
        target (nn.Module): the target network
        tau (float, optional): the soft update factor used to slowly update
            the target network
    """
    for target_param, param in zip(target.parameters(), source.parameters()):
        target_param.data.copy_(tau * param.data + (1.0 - tau) * target_param.data)


def hard_update(source: nn.Module, target: nn.Module) -> None:
    """
    Performs a hard parameter update between two PyTorch Networks.

    Parameters:
        source (nn.Module): the source network
        target (nn.Module): the target network
    """
    for target_param, param in zip(target.parameters(), source.parameters()):
        target_param.data.copy_(param.data)



from typing import List, Tuple

import torch
import torch.nn as nn

from velora.models.activation import ActivationEnum, ActivationTypeLiteral


class MLP(nn.Module):
    """
    A dynamic multi-layer perceptron architecture for feature extraction.

    Parameters:
        in_features (int): the number of input features
        n_hidden (List[int] | int): a list of hidden node sizes or
            a single hidden node size. Dynamically creates Linear layers
            based on sizes
        out_features (int): the number of output features
        activation (ActivationTypeLiteral, optional): the type of activation function to use between layers. Default is 'relu'
        dropout_p (float, optional): a dropout probability rate. Default is '0.0'
    """

    def __init__(
        self,
        in_features: int,
        n_hidden: List[int] | int,
        out_features: int,
        *,
        activation: ActivationTypeLiteral = "relu",
        dropout_p: float = 0.0,
    ) -> None:
        super().__init__()

        self.dropout_p = dropout_p
        n_hidden = [n_hidden] if isinstance(n_hidden, int) else n_hidden

        input = nn.Linear(in_features, n_hidden[0])
        h_layers = self.set_hidden_layers(n_hidden, activation)
        output = nn.Linear(n_hidden[-1], out_features)

        self.fc = nn.Sequential(
            input,
            ActivationEnum.get(activation),
            *h_layers,
            output,
        )

    def set_hidden_layers(self, n_hidden: List[int], activation: str) -> nn.ModuleList:
        """
        Dynamically creates the hidden layers with
        activation functions and dropout layers.
        """
        h_layers = nn.ModuleList()

        for i in range(len(n_hidden) - 1):
            layers = [
                nn.Linear(n_hidden[i], n_hidden[i + 1]),
                ActivationEnum.get(activation),
            ]

            if self.dropout_p > 0.0:
                layers.append(nn.Dropout(self.dropout_p))

            h_layers.extend(layers)

        return h_layers

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Perform a forward pass through the network
        without an output activation.

        Parameters:
            x (torch.Tensor): the input tensor
        """
        return self.fc(x)


class BasicCNN(nn.Module):
    def __init__(self, in_channels: int) -> None:
        """
        A CNN backbone from the DQN Nature paper:
            Mnih, Volodymyr, et al.
            "Human-level control through deep reinforcement learning."
            Nature 518.7540 (2015): 529-533.

        Only contains the Convolutional Network with a flattened output.
        Useful for connecting with other architectures such as the
        velora.models.LiquidNCPNetwork.

        Parameters:
            in_channels (int): the number of channels in the input image.
                E.g., `3` for RGB or `1` for grayscale
        """
        super().__init__()
        self.in_channels = in_channels

        self.conv = nn.Sequential(
            nn.Conv2d(in_channels, 32, kernel_size=8, stride=4, padding=0),
            nn.ReLU(),
            nn.Conv2d(32, 64, kernel_size=4, stride=2, padding=0),
            nn.ReLU(),
            nn.Conv2d(64, 64, kernel_size=3, stride=1, padding=0),
            nn.ReLU(),
            nn.Flatten(),
        )

    def out_size(self, dim: Tuple[int, int]) -> int:
        """
        Calculates the size of the convolution output using a dummy input.

        Parameters:
            dim (Tuple[int, int]): the `(height, width)` of a single image
        """
        with torch.no_grad():
            x = torch.zeros(1, self.in_channels, *dim)
            return self.conv(x).size(1)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass through the network.

        Parameters:
            x (torch.Tensor): a batch of images in the shape
                `(batch_size, in_channels, height, width)`.
                - `batch_size` - number of images in the batch
                - `in_channels` - number of colour channels
                - `height` - height of the images
                - `width` - width of the images
        """
        return self.conv(x)
